'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.addMark = addMark;
exports._delete = _delete;
exports.deleteBackward = deleteBackward;
exports.deleteForward = deleteForward;
exports.insertBlock = insertBlock;
exports.insertFragment = insertFragment;
exports.insertInline = insertInline;
exports.insertText = insertText;
exports.setBlock = setBlock;
exports.setInline = setInline;
exports.splitBlock = splitBlock;
exports.splitInline = splitInline;
exports.removeMark = removeMark;
exports.toggleMark = toggleMark;
exports.unwrapBlock = unwrapBlock;
exports.unwrapInline = unwrapInline;
exports.wrapBlock = wrapBlock;
exports.wrapInline = wrapInline;
exports.wrapText = wrapText;

var _normalize = require('../utils/normalize');

var _normalize2 = _interopRequireDefault(_normalize);

var _atRange = require('./at-range');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Add a `mark` to the characters in the current selection.
 *
 * @param {Transform} transform
 * @param {Mark} mark
 * @return {Transform}
 */

function addMark(transform, mark) {
  mark = _normalize2.default.mark(mark);
  var state = transform.state;
  var _state = state;
  var cursorMarks = _state.cursorMarks;
  var document = _state.document;
  var selection = _state.selection;

  // If the selection is collapsed, add the mark to the cursor instead.

  if (selection.isCollapsed) {
    var marks = document.getMarksAtRange(selection);
    state = state.merge({ cursorMarks: marks.add(mark) });
    transform.state = state;
    return transform;
  }

  return (0, _atRange.addMarkAtRange)(transform, selection, mark);
}

/**
 * Delete at the current selection.
 *
 * @param {Transform} transform
 * @return {Transform}
 */

function _delete(transform) {
  var _transform = transform;
  var state = _transform.state;
  var _state2 = state;
  var document = _state2.document;
  var selection = _state2.selection;

  var after = void 0;

  // When collapsed, there's nothing to do.
  if (selection.isCollapsed) return transform;

  // Determine what the selection will be after deleting.
  var _state3 = state;
  var startText = _state3.startText;
  var startKey = selection.startKey;
  var startOffset = selection.startOffset;
  var endKey = selection.endKey;
  var endOffset = selection.endOffset;

  var block = document.getClosestBlock(startText);
  var highest = block.getHighestChild(startText);
  var previous = block.getPreviousSibling(highest);
  var next = block.getNextSibling(highest);

  if (previous && startOffset == 0 && (endKey != startKey || endOffset == startText.length)) {
    if (previous.kind == 'text') {
      if (next && next.kind == 'text') {
        after = selection.merge({
          anchorKey: previous.key,
          anchorOffset: previous.length,
          focusKey: previous.key,
          focusOffset: previous.length
        });
      } else {
        after = selection.collapseToEndOf(previous);
      }
    } else {
      var last = previous.getTexts().last();
      after = selection.collapseToEndOf(last);
    }
  } else {
    after = selection.collapseToStart();
  }

  // Delete and update the selection.
  transform = (0, _atRange.deleteAtRange)(transform, selection);
  state = transform.state;
  state = state.merge({ selection: after });
  transform.state = state;
  return transform;
}

/**
 * Delete backward `n` characters at the current selection.
 *
 * @param {Transform} transform
 * @param {Number} n (optional)
 * @return {Transform}
 */

function deleteBackward(transform) {
  var n = arguments.length <= 1 || arguments[1] === undefined ? 1 : arguments[1];
  var _transform2 = transform;
  var state = _transform2.state;
  var _state4 = state;
  var document = _state4.document;
  var selection = _state4.selection;

  var after = selection;

  // Determine what the selection should be after deleting.
  var startKey = selection.startKey;

  var startNode = document.getDescendant(startKey);

  if (selection.isExpanded) {
    after = selection.collapseToStart();
  } else if (selection.isAtStartOf(document)) {
    after = selection;
  } else if (selection.isAtStartOf(startNode)) {
    var previous = document.getPreviousText(startNode);
    var prevBlock = document.getClosestBlock(previous);
    var prevInline = document.getClosestInline(previous);

    if (prevBlock && prevBlock.isVoid) {
      after = selection;
    } else if (prevInline && prevInline.isVoid) {
      after = selection;
    } else {
      after = selection.collapseToEndOf(previous);
    }
  } else if (selection.isAtEndOf(startNode) && startNode.length == 1) {
    var block = document.getClosestBlock(startKey);
    var highest = block.getHighestChild(startKey);
    var _previous = block.getPreviousSibling(highest);
    var next = block.getNextSibling(highest);

    if (_previous) {
      if (_previous.kind == 'text') {
        if (next && next.kind == 'text') {
          after = selection.merge({
            anchorKey: _previous.key,
            anchorOffset: _previous.length,
            focusKey: _previous.key,
            focusOffset: _previous.length
          });
        } else {
          after = selection.collapseToEndOf(_previous);
        }
      } else {
        var last = _previous.getTexts().last();
        after = selection.collapseToEndOf(last);
      }
    } else {
      after = selection.moveBackward(n);
    }
  } else {
    after = selection.moveBackward(n);
  }

  // Delete backward and then update the selection.
  transform = (0, _atRange.deleteBackwardAtRange)(transform, selection, n);
  state = transform.state;
  state = state.merge({ selection: after });
  transform.state = state;
  return transform;
}

/**
 * Delete forward `n` characters at the current selection.
 *
 * @param {Transform} transform
 * @param {Number} n (optional)
 * @return {Transform}
 */

function deleteForward(transform) {
  var n = arguments.length <= 1 || arguments[1] === undefined ? 1 : arguments[1];
  var _transform3 = transform;
  var state = _transform3.state;
  var _state5 = state;
  var document = _state5.document;
  var selection = _state5.selection;
  var startText = _state5.startText;
  var startKey = selection.startKey;
  var startOffset = selection.startOffset;

  var after = selection;

  // Determine what the selection should be after deleting.
  var block = document.getClosestBlock(startKey);
  var inline = document.getClosestInline(startKey);
  var highest = block.getHighestChild(startKey);
  var previous = block.getPreviousSibling(highest);
  var next = block.getNextSibling(highest);

  if (selection.isExpanded) {
    after = selection.collapseToStart();
  } else if (block && block.isVoid || inline && inline.isVoid) {
    var nextText = document.getNextText(startKey);
    var prevText = document.getPreviousText(startKey);
    after = next ? selection.collapseToStartOf(nextText) : selection.collapseToEndOf(prevText);
  } else if (previous && startOffset == 0 && startText.length == 1) {
    if (previous.kind == 'text') {
      if (next && next.kind == 'text') {
        after = selection.merge({
          anchorKey: previous.key,
          anchorOffset: previous.length,
          focusKey: previous.key,
          focusOffset: previous.length
        });
      } else {
        after = selection.collapseToEndOf(previous);
      }
    } else {
      var last = previous.getTexts().last();
      after = selection.collapseToEndOf(last);
    }
  }

  // Delete forward and then update the selection.
  transform = (0, _atRange.deleteForwardAtRange)(transform, selection, n);
  state = transform.state;
  state = state.merge({ selection: after });
  transform.state = state;
  return transform;
}

/**
 * Insert a `block` at the current selection.
 *
 * @param {Transform} transform
 * @param {String || Object || Block} block
 * @return {Transform}
 */

function insertBlock(transform, block) {
  var _transform4 = transform;
  var state = _transform4.state;
  var _state6 = state;
  var document = _state6.document;
  var selection = _state6.selection;

  var keys = document.getTexts().map(function (text) {
    return text.key;
  });

  // Insert the block
  transform = (0, _atRange.insertBlockAtRange)(transform, selection, block);
  state = transform.state;
  document = state.document;
  selection = state.selection;

  // Determine what the selection should be after inserting.
  var text = document.getTexts().find(function (n) {
    return !keys.includes(n.key);
  });
  selection = selection.collapseToEndOf(text);

  // Update the document and selection.
  state = state.merge({ selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Insert a `fragment` at the current selection.
 *
 * @param {Transform} transform
 * @param {Document} fragment
 * @return {Transform}
 */

function insertFragment(transform, fragment) {
  var _transform5 = transform;
  var state = _transform5.state;
  var _state7 = state;
  var document = _state7.document;
  var selection = _state7.selection;

  var after = selection;

  // If there's nothing in the fragment, do nothing.
  if (!fragment.length) return transform;

  // Lookup some nodes for determining the selection next.
  var lastText = fragment.getTexts().last();
  var lastInline = fragment.getClosestInline(lastText);
  var beforeTexts = document.getTexts();

  // Insert the fragment.
  transform = (0, _atRange.insertFragmentAtRange)(transform, selection, fragment);
  state = transform.state;
  document = state.document;
  selection = state.selection;

  // Determine what the selection should be after inserting.
  var keys = beforeTexts.map(function (text) {
    return text.key;
  });
  var text = document.getTexts().findLast(function (n) {
    return !keys.includes(n.key);
  });
  var previousText = text ? document.getPreviousText(text) : null;

  if (text && lastInline && previousText) {
    after = selection.collapseToEndOf(previousText);
  } else if (text && lastInline) {
    after = selection.collapseToStart();
  } else if (text) {
    after = selection.collapseToStartOf(text).moveForward(lastText.length);
  } else {
    after = selection.collapseToStart().moveForward(lastText.length);
  }

  // Update the document and selection.
  selection = after;
  state = state.merge({ document: document, selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Insert a `inline` at the current selection.
 *
 * @param {Transform} transform
 * @param {String || Object || Block} inline
 * @return {Transform}
 */

function insertInline(transform, inline) {
  var _transform6 = transform;
  var state = _transform6.state;
  var _state8 = state;
  var document = _state8.document;
  var selection = _state8.selection;
  var startText = _state8.startText;

  var hasVoid = document.hasVoidParent(startText);
  var keys = document.getTexts().map(function (text) {
    return text.key;
  });

  // Insert the inline
  transform = (0, _atRange.insertInlineAtRange)(transform, selection, inline);
  state = transform.state;
  document = state.document;
  selection = state.selection;

  // Determine what the selection should be after inserting.
  if (hasVoid) {
    selection = selection;
  } else {
    var text = document.getTexts().find(function (n) {
      return !keys.includes(n.key);
    });
    selection = selection.collapseToEndOf(text);
  }

  // Update the document and selection.
  state = state.merge({ document: document, selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Insert a `text` string at the current selection.
 *
 * @param {Transform} transform
 * @param {String} text
 * @param {Set} marks (optional)
 * @return {Transform}
 */

function insertText(transform, text, marks) {
  var state = transform.state;
  var _state9 = state;
  var cursorMarks = _state9.cursorMarks;
  var document = _state9.document;
  var selection = _state9.selection;

  var after = void 0;
  var isVoid = document.hasVoidParent(state.startText);

  // Determine what the selection should be after inserting.
  if (isVoid) {
    after = selection;
  } else if (selection.isExpanded) {
    after = selection.collapseToStart().moveForward(text.length);
  } else {
    after = selection.moveForward(text.length);
  }

  // Insert the text and update the selection.
  state = (0, _atRange.insertTextAtRange)(transform, selection, text, marks || cursorMarks);
  state = transform.state;
  state = state.merge({ selection: after });
  transform.state = state;
  return transform;
}

/**
 * Set `properties` of the block nodes in the current selection.
 *
 * @param {Transform} transform
 * @param {Object} properties
 * @return {Transform}
 */

function setBlock(transform, properties) {
  var state = transform.state;
  var selection = state.selection;

  return (0, _atRange.setBlockAtRange)(transform, selection, properties);
}

/**
 * Set `properties` of the inline nodes in the current selection.
 *
 * @param {Transform} transform
 * @param {Object} properties
 * @return {Transform}
 */

function setInline(transform, properties) {
  var state = transform.state;
  var selection = state.selection;

  return (0, _atRange.setInlineAtRange)(transform, selection, properties);
}

/**
 * Split the block node at the current selection, to optional `depth`.
 *
 * @param {Transform} transform
 * @param {Number} depth (optional)
 * @return {Transform}
 */

function splitBlock(transform) {
  var depth = arguments.length <= 1 || arguments[1] === undefined ? 1 : arguments[1];
  var _transform7 = transform;
  var state = _transform7.state;

  transform = (0, _atRange.splitBlockAtRange)(transform, state.selection, depth);
  state = transform.state;
  var _state10 = state;
  var document = _state10.document;
  var selection = _state10.selection;

  // Determine what the selection should be after splitting.

  var _selection = selection;
  var startKey = _selection.startKey;

  var startNode = document.getDescendant(startKey);
  var nextNode = document.getNextText(startNode);
  selection = selection.collapseToStartOf(nextNode);
  state = state.merge({ selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Split the inline nodes at the current selection, to optional `depth`.
 *
 * @param {Transform} transform
 * @param {Number} depth (optional)
 * @return {Transform}
 */

function splitInline(transform) {
  var depth = arguments.length <= 1 || arguments[1] === undefined ? Infinity : arguments[1];
  var _transform8 = transform;
  var state = _transform8.state;
  var _state11 = state;
  var document = _state11.document;
  var selection = _state11.selection;

  // Split the document.

  transform = (0, _atRange.splitInlineAtRange)(transform, selection, depth);
  state = transform.state;
  document = state.document;
  selection = state.selection;

  // Determine what the selection should be after splitting.
  var _selection2 = selection;
  var startKey = _selection2.startKey;

  var inlineParent = document.getClosestInline(startKey);

  if (inlineParent) {
    var startNode = document.getDescendant(startKey);
    var nextNode = document.getNextText(startNode);
    selection = selection.collapseToStartOf(nextNode);
  }

  state = state.merge({ document: document, selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Remove a `mark` from the characters in the current selection.
 *
 * @param {Transform} transform
 * @param {Mark} mark
 * @return {Transform}
 */

function removeMark(transform, mark) {
  mark = _normalize2.default.mark(mark);
  var state = transform.state;
  var _state12 = state;
  var cursorMarks = _state12.cursorMarks;
  var document = _state12.document;
  var selection = _state12.selection;

  // If the selection is collapsed, remove the mark from the cursor instead.

  if (selection.isCollapsed) {
    var marks = document.getMarksAtRange(selection);
    state = state.merge({ cursorMarks: marks.remove(mark) });
    transform.state = state;
    return transform;
  }

  return (0, _atRange.removeMarkAtRange)(transform, state.selection, mark);
}

/**
 * Add or remove a `mark` from the characters in the current selection,
 * depending on whether it's already there.
 *
 * @param {Transform} transform
 * @param {Mark} mark
 * @return {Transform}
 */

function toggleMark(transform, mark) {
  mark = _normalize2.default.mark(mark);
  var state = transform.state;

  var exists = state.marks.some(function (m) {
    return m.equals(mark);
  });
  return exists ? removeMark(transform, mark) : addMark(transform, mark);
}

/**
 * Unwrap the current selection from a block parent with `properties`.
 *
 * @param {Transform} transform
 * @param {Object or String} properties
 * @return {Transform}
 */

function unwrapBlock(transform, properties) {
  var state = transform.state;
  var selection = state.selection;

  return (0, _atRange.unwrapBlockAtRange)(transform, selection, properties);
}

/**
 * Unwrap the current selection from an inline parent with `properties`.
 *
 * @param {Transform} transform
 * @param {Object or String} properties
 * @return {Transform}
 */

function unwrapInline(transform, properties) {
  var state = transform.state;
  var selection = state.selection;

  return (0, _atRange.unwrapInlineAtRange)(transform, selection, properties);
}

/**
 * Wrap the block nodes in the current selection with a new block node with
 * `properties`.
 *
 * @param {Transform} transform
 * @param {Object or String} properties
 * @return {Transform}
 */

function wrapBlock(transform, properties) {
  var state = transform.state;
  var selection = state.selection;

  return (0, _atRange.wrapBlockAtRange)(transform, selection, properties);
}

/**
 * Wrap the current selection in new inline nodes with `properties`.
 *
 * @param {Transform} transform
 * @param {Object or String} properties
 * @return {Transform}
 */

function wrapInline(transform, properties) {
  var _transform9 = transform;
  var state = _transform9.state;
  var _state13 = state;
  var document = _state13.document;
  var selection = _state13.selection;
  var _selection3 = selection;
  var startKey = _selection3.startKey;

  var previous = document.getPreviousText(startKey);

  transform = (0, _atRange.wrapInlineAtRange)(transform, selection, properties);
  state = transform.state;
  document = state.document;
  selection = state.selection;

  // Determine what the selection should be after wrapping.
  if (selection.isCollapsed) {
    selection = selection;
  } else if (selection.startOffset == 0) {
    var text = previous ? document.getNextText(previous) : document.getTexts().first();
    selection = selection.moveToRangeOf(text);
  } else if (selection.startKey == selection.endKey) {
    var _text = document.getNextText(selection.startKey);
    selection = selection.moveToRangeOf(_text);
  } else {
    var anchor = document.getNextText(selection.anchorKey);
    var focus = document.getDescendant(selection.focusKey);
    selection = selection.merge({
      anchorKey: anchor.key,
      anchorOffset: 0,
      focusKey: focus.key,
      focusOffset: selection.focusOffset
    });
  }

  state = state.merge({ selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Wrap the current selection with prefix/suffix.
 *
 * @param {Transform} transform
 * @param {String} prefix
 * @param {String} suffix
 * @return {Transform}
 */

function wrapText(transform, prefix) {
  var suffix = arguments.length <= 2 || arguments[2] === undefined ? prefix : arguments[2];
  var _transform10 = transform;
  var state = _transform10.state;
  var _state14 = state;
  var document = _state14.document;
  var selection = _state14.selection;
  var anchorOffset = selection.anchorOffset;
  var anchorKey = selection.anchorKey;
  var focusOffset = selection.focusOffset;
  var focusKey = selection.focusKey;
  var isBackward = selection.isBackward;

  var after = void 0;

  // Determine what the selection should be after wrapping.
  if (anchorKey == focusKey) {
    after = selection.moveForward(prefix.length);
  } else {
    after = selection.merge({
      anchorOffset: isBackward ? anchorOffset : anchorOffset + prefix.length,
      focusOffset: isBackward ? focusOffset + prefix.length : focusOffset
    });
  }

  // Wrap the text and update the state.
  transform = (0, _atRange.wrapTextAtRange)(transform, selection, prefix, suffix);
  state = transform.state;
  state = state.merge({ selection: after });
  transform.state = state;
  return transform;
}