'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.addMarkAtRange = addMarkAtRange;
exports.deleteAtRange = deleteAtRange;
exports.deleteBackwardAtRange = deleteBackwardAtRange;
exports.deleteForwardAtRange = deleteForwardAtRange;
exports.insertBlockAtRange = insertBlockAtRange;
exports.insertFragmentAtRange = insertFragmentAtRange;
exports.insertInlineAtRange = insertInlineAtRange;
exports.insertTextAtRange = insertTextAtRange;
exports.removeMarkAtRange = removeMarkAtRange;
exports.setBlockAtRange = setBlockAtRange;
exports.setInlineAtRange = setInlineAtRange;
exports.splitBlockAtRange = splitBlockAtRange;
exports.splitInlineAtRange = splitInlineAtRange;
exports.toggleMarkAtRange = toggleMarkAtRange;
exports.unwrapBlockAtRange = unwrapBlockAtRange;
exports.unwrapInlineAtRange = unwrapInlineAtRange;
exports.wrapBlockAtRange = wrapBlockAtRange;
exports.wrapInlineAtRange = wrapInlineAtRange;
exports.wrapTextAtRange = wrapTextAtRange;

var _block = require('../models/block');

var _block2 = _interopRequireDefault(_block);

var _inline = require('../models/inline');

var _inline2 = _interopRequireDefault(_inline);

var _normalize = require('../utils/normalize');

var _normalize2 = _interopRequireDefault(_normalize);

var _selection = require('../models/selection');

var _selection2 = _interopRequireDefault(_selection);

var _text = require('../models/text');

var _text2 = _interopRequireDefault(_text);

var _isInRange = require('../utils/is-in-range');

var _isInRange2 = _interopRequireDefault(_isInRange);

var _uid = require('../utils/uid');

var _uid2 = _interopRequireDefault(_uid);

var _immutable = require('immutable');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Add a new `mark` to the characters at `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Mixed} mark
 * @return {Transform}
 */

function addMarkAtRange(transform, range, mark) {
  if (range.isCollapsed) return transform;

  var state = transform.state;
  var document = state.document;
  var startKey = range.startKey;
  var startOffset = range.startOffset;
  var endKey = range.endKey;
  var endOffset = range.endOffset;

  var texts = document.getTextsAtRange(range);

  texts.forEach(function (text) {
    var key = text.key;

    var index = 0;
    var length = text.length;

    if (key == startKey) index = startOffset;
    if (key == endKey) length = endOffset;
    if (key == startKey && key == endKey) length = endOffset - startOffset;

    transform.addMarkByKey(key, index, length, mark);
  });

  return transform;
}

/**
 * Delete everything in a `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @return {Transform}
 */

function deleteAtRange(transform, range) {
  // if (range.isCollapsed) return transform

  // const { state } = transform
  // const { document } = state
  // const { startKey, startOffset, endKey, endOffset } = range
  // const startText = document.assertDescendant(startKey)
  // const endText = document.assertDescendant(endKey)

  // if (startKey == endKey) {
  //   const index = startOffset
  //   const length = endOffset - startOffset
  //   return transform.removeTextByKey(startKey, index, length)
  // }

  var _transform = transform;
  var state = _transform.state;
  var _state = state;
  var document = _state.document;

  // If the range is collapsed, there's nothing to delete.

  if (range.isCollapsed) return transform;

  // Make sure the children exist.
  var startKey = range.startKey;
  var startOffset = range.startOffset;
  var endKey = range.endKey;
  var endOffset = range.endOffset;

  document.assertDescendant(startKey);
  document.assertDescendant(endKey);

  // If the start and end nodes are the same, just remove characters.
  if (startKey == endKey) {
    var text = document.getDescendant(startKey);
    text = text.removeText(startOffset, endOffset - startOffset);
    document = document.updateDescendant(text);
    document = document.normalize();
    state = state.merge({ document: document });
    transform.state = state;
    return transform;
  }

  // Split the blocks and determine the edge boundaries.
  var start = range.collapseToStart();
  var end = range.collapseToEnd();
  var startBlock = document.getClosestBlock(startKey);
  var endBlock = document.getClosestBlock(endKey);
  var startDepth = document.getDepth(startBlock);
  var endDepth = document.getDepth(endBlock);

  var ancestor = document.getCommonAncestor(startKey, endKey);
  var isAncestor = ancestor == document;
  var ancestorDepth = isAncestor ? 0 : document.getDepth(ancestor);

  transform = splitBlockAtRange(transform, start, startDepth - ancestorDepth);
  transform = splitBlockAtRange(transform, end, endDepth - ancestorDepth);
  state = transform.state;
  document = state.document;
  ancestor = document.getCommonAncestor(startKey, endKey);

  var startText = ancestor.getDescendant(startKey);
  var startEdgeText = ancestor.getNextText(startKey);

  var endText = ancestor.getNextText(endKey);
  var endEdgeText = ancestor.getDescendant(endKey);

  startBlock = document.getClosestBlock(startText);
  endBlock = document.getClosestBlock(endText);

  // Remove the new blocks inside the edges.
  var startEdgeBlock = ancestor.getFurthestBlock(startEdgeText);
  var endEdgeBlock = ancestor.getFurthestBlock(endEdgeText);

  var nodes = ancestor.nodes.takeUntil(function (n) {
    return n == startEdgeBlock;
  }).concat(ancestor.nodes.skipUntil(function (n) {
    return n == endEdgeBlock;
  }).rest());

  ancestor = ancestor.merge({ nodes: nodes });

  // Take the end edge's inline nodes and move them to the start edge.
  var startNodes = startBlock.nodes.concat(endBlock.nodes);
  startBlock = startBlock.merge({ nodes: startNodes });
  ancestor = ancestor.updateDescendant(startBlock);

  // While the end child is an only child, remove the block it's in.
  var endParent = ancestor.getClosestBlock(endBlock);

  while (endParent && endParent.nodes.size == 1) {
    endBlock = endParent;
    endParent = ancestor.getClosestBlock(endParent);
  }

  ancestor = ancestor.removeDescendant(endBlock);

  // Update the document.
  document = isAncestor ? ancestor : document.updateDescendant(ancestor);

  // Normalize the adjacent text nodes.
  document = document.normalize();

  // Update the state.
  state = state.merge({ document: document });
  transform.state = state;
  return transform;
}

/**
 * Delete backward `n` characters at a `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Number} n (optional)
 * @return {Transform}
 */

function deleteBackwardAtRange(transform, range) {
  var n = arguments.length <= 2 || arguments[2] === undefined ? 1 : arguments[2];
  var state = transform.state;
  var document = state.document;
  var _range = range;
  var startKey = _range.startKey;
  var focusOffset = _range.focusOffset;

  var text = document.getDescendant(startKey);
  var block = document.getClosestBlock(startKey);
  var inline = document.getClosestInline(startKey);

  if (range.isExpanded) {
    return transform.deleteAtRange(range);
  }

  if (block && block.isVoid) {
    return transform.removeNodeByKey(block.key);
  }

  if (inline && inline.isVoid) {
    return transform.removeNodeByKey(inline.key);
  }

  if (range.isAtStartOf(document)) {
    return transform;
  }

  if (range.isAtStartOf(text)) {
    var prev = document.getPreviousText(text);
    var prevBlock = document.getClosestBlock(prev);
    var prevInline = document.getClosestInline(prev);

    if (prevBlock && prevBlock.isVoid) {
      return transform.removeNodeByKey(prevBlock.key);
    }

    if (prevInline && prevInline.isVoid) {
      return transform.removeNodeByKey(prevInline.key);
    }

    range = range.merge({
      anchorKey: prev.key,
      anchorOffset: prev.length
    });

    return transform.deleteAtRange(range);
  }

  range = range.merge({
    focusOffset: focusOffset - 1,
    isBackward: true
  });

  return transform.deleteAtRange(range);
}

/**
 * Delete forward `n` characters at a `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Number} n (optional)
 * @return {Transform}
 */

function deleteForwardAtRange(transform, range) {
  var n = arguments.length <= 2 || arguments[2] === undefined ? 1 : arguments[2];
  var state = transform.state;
  var document = state.document;
  var _range2 = range;
  var startKey = _range2.startKey;
  var focusOffset = _range2.focusOffset;

  var text = document.getDescendant(startKey);
  var inline = document.getClosestInline(startKey);
  var block = document.getClosestBlock(startKey);

  if (range.isExpanded) {
    return transform.deleteAtRange(range);
  }

  if (block && block.isVoid) {
    return transform.removeNodeByKey(block.key);
  }

  if (inline && inline.isVoid) {
    return transform.removeNodeByKey(inline.key);
  }

  if (range.isAtEndOf(document)) {
    return transform;
  }

  if (range.isAtEndOf(text)) {
    var next = document.getNextText(text);

    range = range.merge({
      focusKey: next.key,
      focusOffset: 0
    });

    return transform.deleteAtRange(range);
  }

  range = range.merge({
    focusOffset: focusOffset + 1
  });

  return transform.deleteAtRange(range);
}

/**
 * Insert a `block` node at `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Block or String or Object} block
 * @return {Transform}
 */

function insertBlockAtRange(transform, range, block) {
  block = _normalize2.default.block(block);

  if (range.isExpanded) {
    transform.deleteAtRange(range);
    range = range.collapseToStart();
  }

  var state = transform.state;
  var document = state.document;
  var _range3 = range;
  var startKey = _range3.startKey;
  var startOffset = _range3.startOffset;

  var startText = document.assertDescendant(startKey);
  var startBlock = document.getClosestBlock(startKey);
  var parent = document.getParent(startBlock);
  var index = parent.nodes.indexOf(startBlock);

  if (startBlock.isVoid) {
    transform.insertNodeByKey(parent.key, index + 1, block);
  } else if (startBlock.isEmpty) {
    transform.removeNodeByKey(startBlock.key);
    transform.insertNodeByKey(parent.key, index, block);
  } else if (range.isAtStartOf(startBlock)) {
    transform.insertNodeByKey(parent.key, index, block);
  } else if (range.isAtEndOf(startBlock)) {
    transform.insertNodeByKey(parent.key, index + 1, block);
  } else {
    var offset = startBlock.getOffset(startText) + startOffset;
    transform.splitNodeByKey(startBlock.key, offset);
    transform.insertNodeByKey(parent.key, index + 1, block);
  }

  return transform;
}

/**
 * Insert a `fragment` at a `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Document} fragment
 * @return {Transform}
 */

function insertFragmentAtRange(transform, range, fragment) {
  var _transform2 = transform;
  var state = _transform2.state;
  var _state2 = state;
  var document = _state2.document;

  // Ensure that the selection is normalized.

  range = range.normalize(document);

  // If the range is expanded, delete first.
  if (range.isExpanded) {
    transform = deleteAtRange(transform, range);
    state = transform.state;
    document = state.document;
    range = range.collapseToStart();
  }

  // If the fragment is empty, do nothing.
  if (!fragment.length) return transform;

  // Make sure each node in the fragment has a unique key.
  fragment = fragment.mapDescendants(function (child) {
    return child.set('key', (0, _uid2.default)());
  });

  // Split the inlines if need be.
  if (!document.isInlineSplitAtRange(range)) {
    transform = splitInlineAtRange(transform, range);
    state = transform.state;
    document = state.document;
  }

  // Determine the start and next children to insert into.
  var _range4 = range;
  var startKey = _range4.startKey;
  var endKey = _range4.endKey;

  var block = document.getClosestBlock(startKey);
  var start = document.getDescendant(startKey);
  var startChild = void 0;
  var nextChild = void 0;

  if (range.isAtStartOf(document)) {
    nextChild = document.getClosestBlock(document.getTexts().first());
  }

  if (range.isAtStartOf(block)) {
    nextChild = block.getHighestChild(block.getTexts().first());
  } else if (range.isAtStartOf(start)) {
    startChild = block.getHighestChild(block.getPreviousText(start));
    nextChild = block.getNextSibling(startChild);
  } else {
    startChild = block.getHighestChild(start);
    nextChild = block.getNextSibling(startChild);
  }

  // Get the first and last block of the fragment.
  var blocks = fragment.getBlocks();
  var firstBlock = blocks.first();
  var lastBlock = blocks.last();

  // If the block is empty, merge in the first block's type and data.
  if (block.length == 0) {
    block = block.merge({
      type: firstBlock.type,
      data: firstBlock.data
    });
  }

  // Insert the first blocks nodes into the starting block.
  if (startChild) {
    block = block.insertChildrenAfter(startChild, firstBlock.nodes);
  } else {
    block = block.insertChildrenBefore(nextChild, firstBlock.nodes);
  }

  document = document.updateDescendant(block);

  // If there are no other siblings, that's it.
  if (firstBlock == lastBlock) {
    document = document.normalize();
    state = state.merge({ document: document });
    transform.state = state;
    return transform;
  }

  // Otherwise, remove the fragment's first block's highest solo parent...
  var highestParent = fragment.getHighestOnlyChildParent(firstBlock);
  fragment = fragment.removeDescendant(highestParent || firstBlock);

  // Then, add the inlines after the cursor from the current block to the
  // start of the last block in the fragment.
  if (nextChild) {
    lastBlock = lastBlock.concatChildren(block.getChildrenAfterIncluding(nextChild));
    fragment = fragment.updateDescendant(lastBlock);

    block = block.removeChildrenAfterIncluding(nextChild);
    document = document.updateDescendant(block);
  }

  // Finally, add the fragment's children after the block.
  document = document.insertChildrenAfter(block, fragment.nodes);
  document = document.normalize();
  state = state.merge({ document: document });
  transform.state = state;
  return transform;
}

/**
 * Insert an `inline` node at `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Inline or String or Object} inline
 * @return {Transform}
 */

function insertInlineAtRange(transform, range, inline) {
  inline = _normalize2.default.inline(inline);

  if (range.isExpanded) {
    transform.deleteAtRange(range);
    range = range.collapseToStart();
  }

  var state = transform.state;
  var document = state.document;
  var _range5 = range;
  var startKey = _range5.startKey;
  var startOffset = _range5.startOffset;

  var parent = document.getParent(startKey);
  var startText = document.assertDescendant(startKey);
  var index = parent.nodes.indexOf(startText);

  if (parent.isVoid) {
    return transform;
  }

  transform.splitNodeByKey(startKey, startOffset);
  transform.insertNodeByKey(parent.key, index + 1, inline);
  return transform;
}

/**
 * Insert `text` at a `range`, with optional `marks`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {String} text
 * @param {Set} marks (optional)
 * @return {Transform}
 */

function insertTextAtRange(transform, range, text, marks) {
  var state = transform.state;
  var document = state.document;
  var startKey = range.startKey;
  var startOffset = range.startOffset;

  var parent = document.getParent(startKey);

  if (parent.isVoid) {
    return transform;
  }

  if (range.isExpanded) {
    transform.deleteAtRange(range);
  }

  transform.insertTextByKey(startKey, startOffset, text, marks);
  return transform;
}

/**
 * Remove an existing `mark` to the characters at `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Mark or String} mark (optional)
 * @return {Transform}
 */

function removeMarkAtRange(transform, range, mark) {
  if (range.isCollapsed) return transform;

  var state = transform.state;
  var document = state.document;

  var texts = document.getTextsAtRange(range);
  var startKey = range.startKey;
  var startOffset = range.startOffset;
  var endKey = range.endKey;
  var endOffset = range.endOffset;


  texts.forEach(function (text) {
    var key = text.key;

    var index = 0;
    var length = text.length;

    if (key == startKey) index = startOffset;
    if (key == endKey) length = endOffset;
    if (key == startKey && key == endKey) length = endOffset - startOffset;

    transform.removeMarkByKey(key, index, length, mark);
  });

  return transform;
}

/**
 * Set the `properties` of block nodes in a `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Object || String} properties
 * @return {Transform}
 */

function setBlockAtRange(transform, range, properties) {
  var state = transform.state;
  var document = state.document;

  var blocks = document.getBlocksAtRange(range);

  blocks.forEach(function (block) {
    transform.setNodeByKey(block.key, properties);
  });

  return transform;
}

/**
 * Set the `properties` of inline nodes in a `range`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Object || String} properties
 * @return {Transform}
 */

function setInlineAtRange(transform, range, properties) {
  var state = transform.state;
  var document = state.document;

  var inlines = document.getInlinesAtRange(range);

  inlines.forEach(function (inline) {
    transform.setNodeByKey(inline.key, properties);
  });

  return transform;
}

/**
 * Split the block nodes at a `range`, to optional `height`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Number} height (optional)
 * @return {Transform}
 */

function splitBlockAtRange(transform, range) {
  var height = arguments.length <= 2 || arguments[2] === undefined ? 1 : arguments[2];

  if (range.isExpanded) {
    transform.deleteAtRange(range);
    range = range.collapseToStart();
  }

  var _range6 = range;
  var startKey = _range6.startKey;
  var startOffset = _range6.startOffset;
  var state = transform.state;
  var document = state.document;

  var node = document.assertDescendant(startKey);
  var parent = document.getClosestBlock(node);
  var offset = startOffset;
  var h = 0;

  while (parent && parent.kind == 'block' && h < height) {
    offset += parent.getOffset(node);
    node = parent;
    parent = document.getClosestBlock(parent);
    h++;
  }

  transform.splitNodeByKey(node.key, offset);
  transform.normalizeDocument();
  return transform;
}

/**
 * Split the inline nodes at a `range`, to optional `height`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Number} height (optiona)
 * @return {Transform}
 */

function splitInlineAtRange(transform, range) {
  var height = arguments.length <= 2 || arguments[2] === undefined ? Infinity : arguments[2];

  if (range.isExpanded) {
    transform.deleteAtRange(range);
    range = range.collapseToStart();
  }

  var _range7 = range;
  var startKey = _range7.startKey;
  var startOffset = _range7.startOffset;
  var state = transform.state;
  var document = state.document;

  var node = document.assertDescendant(startKey);
  var parent = document.getClosestInline(node);
  var offset = startOffset;
  var h = 0;

  while (parent && parent.kind == 'inline' && h < height) {
    offset += parent.getOffset(node);
    node = parent;
    parent = document.getClosestInline(parent);
    h++;
  }

  return transform.splitNodeByKey(node.key, offset);
}

/**
 * Add or remove a `mark` from the characters at `range`, depending on whether
 * it's already there.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {Mixed} mark
 * @return {Transform}
 */

function toggleMarkAtRange(transform, range, mark) {
  if (range.isCollapsed) return transform;

  mark = _normalize2.default.mark(mark);

  var state = transform.state;
  var document = state.document;

  var marks = document.getMarksAtRange(range);
  var exists = marks.some(function (m) {
    return m.equals(mark);
  });

  if (exists) {
    transform.removeMarkAtRange(range, mark);
  } else {
    transform.addMarkAtRange(range, mark);
  }

  return transform;
}

/**
 * Unwrap all of the block nodes in a `range` from a block with `properties`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {String or Object} properties
 * @return {Transform}
 */

function unwrapBlockAtRange(transform, range, properties) {
  var state = transform.state;

  properties = _normalize2.default.nodeProperties(properties);
  var _state3 = state;
  var document = _state3.document;

  // Get the deepest blocks in the range.

  var blocks = document.getBlocksAtRange(range);

  // Get the matching wrapper blocks.
  var wrappers = blocks.reduce(function (memo, text) {
    var match = document.getClosest(text, function (parent) {
      if (parent.kind != 'block') return false;
      if (properties.type && parent.type != properties.type) return false;
      if (properties.data && !parent.data.isSuperset(properties.data)) return false;
      return true;
    });

    if (match) memo = memo.add(match);
    return memo;
  }, new _immutable.Set());

  // For each of the wrapper blocks...
  wrappers.forEach(function (wrapper) {
    var first = wrapper.nodes.first();
    var last = wrapper.nodes.last();
    var parent = document.getParent(wrapper);

    // Get the wrapped direct children.
    var children = wrapper.nodes.filter(function (child) {
      return blocks.some(function (block) {
        return child == block || child.hasDescendant(block);
      });
    });

    // Determine what the new nodes should be...
    var firstMatch = children.first();
    var lastMatch = children.last();
    var nodes = void 0;

    // If the first and last both match, remove the wrapper completely.
    if (first == firstMatch && last == lastMatch) {
      nodes = parent.nodes.takeUntil(function (n) {
        return n == wrapper;
      }).concat(wrapper.nodes).concat(parent.nodes.skipUntil(function (n) {
        return n == wrapper;
      }).rest());
    }

    // If only the last child matches, move the last nodes.
    else if (last == lastMatch) {
        var remain = wrapper.nodes.takeUntil(function (n) {
          return n == firstMatch;
        });
        var updated = wrapper.merge({ nodes: remain });
        nodes = parent.nodes.takeUntil(function (n) {
          return n == wrapper;
        }).push(updated).concat(children).concat(parent.nodes.skipUntil(function (n) {
          return n == wrapper;
        }).rest());
      }

      // If only the first child matches, move the first ones.
      else if (first == firstMatch) {
          var _remain = wrapper.nodes.skipUntil(function (n) {
            return n == lastMatch;
          }).rest();
          var _updated = wrapper.merge({ nodes: _remain });
          nodes = parent.nodes.takeUntil(function (n) {
            return n == wrapper;
          }).concat(children).push(_updated).concat(parent.nodes.skipUntil(function (n) {
            return n == wrapper;
          }).rest());
        }

        // Otherwise, move the middle ones.
        else {
            var firsts = wrapper.nodes.takeUntil(function (n) {
              return n == firstMatch;
            });
            var lasts = wrapper.nodes.skipUntil(function (n) {
              return n == lastMatch;
            }).rest();
            var updatedFirst = wrapper.merge({ nodes: firsts });
            var updatedLast = wrapper.merge({ nodes: lasts });
            nodes = parent.nodes.takeUntil(function (n) {
              return n == wrapper;
            }).push(updatedFirst).concat(children).push(updatedLast).concat(parent.nodes.skipUntil(function (n) {
              return n == wrapper;
            }).rest());
          }

    document = parent == document ? document.merge({ nodes: nodes }) : document.updateDescendant(parent.merge({ nodes: nodes }));
  });

  document = document.normalize();
  state = state.merge({ document: document });
  transform.state = state;
  return transform;
}

/**
 * Unwrap the inline nodes in a `range` from an inline with `properties`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {String or Object} properties
 * @return {Transform}
 */

function unwrapInlineAtRange(transform, range, properties) {
  var state = transform.state;

  properties = _normalize2.default.nodeProperties(properties);
  var _state4 = state;
  var document = _state4.document;

  var blocks = document.getInlinesAtRange(range);

  // Find the closest matching inline wrappers of each text node.
  var texts = document.getTexts();
  var wrappers = texts.reduce(function (memo, text) {
    var match = document.getClosest(text, function (parent) {
      if (parent.kind != 'inline') return false;
      if (properties.type && parent.type != properties.type) return false;
      if (properties.data && !parent.data.isSuperset(properties.data)) return false;
      return true;
    });

    if (match) memo = memo.add(match);
    return memo;
  }, new _immutable.Set());

  // Replace each of the wrappers with their child nodes.
  wrappers.forEach(function (wrapper) {
    var parent = document.getParent(wrapper);

    // Replace the wrapper in the parent's nodes with the block.
    var nodes = parent.nodes.takeUntil(function (n) {
      return n == wrapper;
    }).concat(wrapper.nodes).concat(parent.nodes.skipUntil(function (n) {
      return n == wrapper;
    }).rest());

    // Update the parent.
    document = parent == document ? document.merge({ nodes: nodes }) : document.updateDescendant(parent.merge({ nodes: nodes }));
  });

  document = document.normalize();
  state = state.merge({ document: document });
  transform.state = state;
  return transform;
}

/**
 * Wrap all of the blocks in a `range` in a new block with `properties`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {String or Object} properties
 * @return {Transform}
 */

function wrapBlockAtRange(transform, range, properties) {
  var state = transform.state;

  properties = _normalize2.default.nodeProperties(properties);
  var _state5 = state;
  var document = _state5.document;

  // Get the block nodes, sorted by depth.

  var blocks = document.getBlocksAtRange(range);
  var sorted = blocks.sort(function (a, b) {
    var da = document.getDepth(a);
    var db = document.getDepth(b);
    if (da == db) return 0;else if (da > db) return -1;else return 1;
  });

  // Get the lowest common siblings, relative to the highest block.
  var highest = sorted.first();
  var depth = document.getDepth(highest);
  var siblings = blocks.reduce(function (memo, block) {
    var sibling = document.getDepth(block) == depth ? block : document.getClosest(block, function (p) {
      return document.getDepth(p) == depth;
    });
    memo = memo.push(sibling);
    return memo;
  }, _block2.default.createList());

  // Wrap the siblings in a new block.
  var wrapper = _block2.default.create({
    nodes: siblings,
    type: properties.type,
    data: properties.data
  });

  // Replace the siblings with the wrapper.
  var first = siblings.first();
  var last = siblings.last();
  var parent = document.getParent(highest);
  var nodes = parent.nodes.takeUntil(function (n) {
    return n == first;
  }).push(wrapper).concat(parent.nodes.skipUntil(function (n) {
    return n == last;
  }).rest());

  // Update the parent.
  document = parent == document ? document.merge({ nodes: nodes }) : document.updateDescendant(parent.merge({ nodes: nodes }));

  state = state.merge({ document: document });
  transform.state = state;
  return transform;
}

/**
 * Wrap the text and inlines in a `range` in a new inline with `properties`.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {String or Object} properties
 * @return {Transform}
 */

function wrapInlineAtRange(transform, range, properties) {
  var _transform3 = transform;
  var state = _transform3.state;

  properties = _normalize2.default.nodeProperties(properties);
  var _state6 = state;
  var document = _state6.document;

  // If collapsed, there's nothing to wrap.

  if (range.isCollapsed) return transform;

  // Split at the start of the range.
  var start = range.collapseToStart();
  transform = splitInlineAtRange(transform, start);
  state = transform.state;
  document = state.document;

  // Determine the new end of the range, and split there.
  var _range8 = range;
  var startKey = _range8.startKey;
  var startOffset = _range8.startOffset;
  var endKey = _range8.endKey;
  var endOffset = _range8.endOffset;

  var firstNode = document.getDescendant(startKey);
  var nextNode = document.getNextText(startKey);
  var end = startKey != endKey ? range.collapseToEnd() : _selection2.default.create({
    anchorKey: nextNode.key,
    anchorOffset: endOffset - startOffset,
    focusKey: nextNode.key,
    focusOffset: endOffset - startOffset
  });

  transform = splitInlineAtRange(transform, end);
  state = transform.state;
  document = state.document;

  // Calculate the new range to wrap around.
  var endNode = document.getDescendant(end.anchorKey);
  range = _selection2.default.create({
    anchorKey: nextNode.key,
    anchorOffset: 0,
    focusKey: endNode.key,
    focusOffset: endNode.length
  });

  // Get the furthest inline nodes in the range.
  var texts = document.getTextsAtRange(range);
  var children = texts.map(function (text) {
    return document.getFurthestInline(text) || text;
  });

  // Iterate each of the child nodes, wrapping them.
  children.forEach(function (child) {
    var wrapper = _inline2.default.create({
      nodes: [child],
      type: properties.type,
      data: properties.data
    });

    // Replace the child in it's parent with the wrapper.
    var parent = document.getParent(child);
    var nodes = parent.nodes.takeUntil(function (n) {
      return n == child;
    }).push(wrapper).concat(parent.nodes.skipUntil(function (n) {
      return n == child;
    }).rest());

    // Update the parent.
    document = parent == document ? document.merge({ nodes: nodes }) : document.updateDescendant(parent.merge({ nodes: nodes }));
  });

  document = document.normalize();
  state = state.merge({ document: document });
  transform.state = state;
  return transform;
}

/**
 * Wrap the text in a `range` in a prefix/suffix.
 *
 * @param {Transform} transform
 * @param {Selection} range
 * @param {String} prefix
 * @param {String} suffix (optional)
 * @return {Transform}
 */

function wrapTextAtRange(transform, range, prefix) {
  var suffix = arguments.length <= 3 || arguments[3] === undefined ? prefix : arguments[3];
  var startKey = range.startKey;
  var endKey = range.endKey;

  var start = range.collapseToStart();
  var end = range.collapseToEnd();

  if (startKey == endKey) {
    end = end.moveForward(prefix.length);
  }

  transform.insertTextAtRange(start, prefix);
  transform.insertTextAtRange(end, suffix);
  return transform;
}