'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.addMarkByKey = addMarkByKey;
exports.insertNodeByKey = insertNodeByKey;
exports.insertTextByKey = insertTextByKey;
exports.moveNodeByKey = moveNodeByKey;
exports.removeMarkByKey = removeMarkByKey;
exports.removeNodeByKey = removeNodeByKey;
exports.removeTextByKey = removeTextByKey;
exports.setMarkByKey = setMarkByKey;
exports.setNodeByKey = setNodeByKey;
exports.splitNodeByKey = splitNodeByKey;

var _normalize = require('../utils/normalize');

var _normalize2 = _interopRequireDefault(_normalize);

var _uid = require('../utils/uid');

var _uid2 = _interopRequireDefault(_uid);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Add mark to text at `offset` and `length` in node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Number} offset
 * @param {Number} length
 * @param {Mixed} mark
 * @return {Transform}
 */

function addMarkByKey(transform, key, offset, length, mark) {
  mark = _normalize2.default.mark(mark);

  var state = transform.state;
  var _state = state;
  var document = _state.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);

  node = node.addMark(offset, length, mark);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'add-mark',
    offset: offset,
    length: length,
    mark: mark,
    path: path
  });

  return transform;
}

/**
 * Insert a `node` at `index` in a node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Number} index
 * @param {Node} node
 * @return {Transform}
 */

function insertNodeByKey(transform, key, index, node) {
  var state = transform.state;
  var _state2 = state;
  var document = _state2.document;

  var parent = document.key == key ? document : document.assertDescendant(key);
  var isParent = document == parent;
  var path = document.getPath(parent);
  var nodes = parent.nodes.splice(index, 0, node);

  parent = parent.merge({ nodes: nodes });
  document = isParent ? parent : document.updateDescendant(parent);
  document = document.normalize();
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'insert-node',
    index: index,
    node: node,
    path: path
  });

  return transform;
}

/**
 * Insert `text` at `offset` in node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Number} offset
 * @param {String} text
 * @param {Set} marks (optional)
 * @return {Transform}
 */

function insertTextByKey(transform, key, offset, text, marks) {
  var state = transform.state;
  var _state3 = state;
  var document = _state3.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);

  node = node.insertText(offset, text, marks);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'insert-text',
    offset: offset,
    marks: marks,
    path: path,
    text: text
  });

  return transform;
}

/**
 * Move a node by `key` to a new parent by `key` and `index`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {String} newKey
 * @param {Number} index
 * @return {Transform}
 */

function moveNodeByKey(transform, key, newKey, newIndex) {
  var state = transform.state;
  var _state4 = state;
  var document = _state4.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);
  var newPath = document.getPath(newKey);
  var parent = document.getParent(node);
  var isParent = document == parent;
  var index = parent.nodes.indexOf(node);

  parent = parent.removeNode(index);
  document = isParent ? parent : document.updateDescendant(parent);

  var target = document.assertDescendant(newKey);
  var isTarget = document == target;

  target = target.insertNode(newIndex, node);
  document = isTarget ? target : document.updateDescendant(target);
  document = document.normalize();
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'move-node',
    path: path,
    newPath: newPath,
    newIndex: newIndex
  });

  return transform;
}

/**
 * Remove mark from text at `offset` and `length` in node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Number} offset
 * @param {Number} length
 * @param {Mark} mark
 * @return {Transform}
 */

function removeMarkByKey(transform, key, offset, length, mark) {
  mark = _normalize2.default.mark(mark);

  var state = transform.state;
  var _state5 = state;
  var document = _state5.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);

  node = node.removeMark(offset, length, mark);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'remove-mark',
    offset: offset,
    length: length,
    mark: mark,
    path: path
  });

  return transform;
}

/**
 * Remove a node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @return {Transform}
 */

function removeNodeByKey(transform, key) {
  var state = transform.state;
  var _state6 = state;
  var document = _state6.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);
  var parent = document.getParent(node);
  var index = parent.nodes.indexOf(node);
  var isParent = document == parent;

  parent = parent.removeNode(index);
  document = isParent ? parent : document.updateDescendant(parent);
  document = document.normalize();
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'remove-node',
    path: path
  });

  return transform;
}

/**
 * Remove text at `offset` and `length` in node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Number} offset
 * @param {Number} length
 * @return {Transform}
 */

function removeTextByKey(transform, key, offset, length) {
  var state = transform.state;
  var _state7 = state;
  var document = _state7.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);

  node = node.removeText(offset, length);
  document = document.updateDescendant(node);
  document = document.normalize();
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'remove-text',
    offset: offset,
    length: length,
    path: path
  });

  return transform;
}

/**
 * Set `properties` on mark on text at `offset` and `length` in node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Number} offset
 * @param {Number} length
 * @param {Mark} mark
 * @return {Transform}
 */

function setMarkByKey(transform, key, offset, length, mark, properties) {
  mark = _normalize2.default.mark(mark);
  properties = _normalize2.default.markProperties(properties);

  var state = transform.state;
  var _state8 = state;
  var document = _state8.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);

  node = node.updateMark(offset, length, mark, properties);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'set-mark',
    offset: offset,
    length: length,
    mark: mark,
    path: path,
    properties: properties
  });

  return transform;
}

/**
 * Set `properties` on a node by `key`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Object || String} properties
 * @return {Transform}
 */

function setNodeByKey(transform, key, properties) {
  properties = _normalize2.default.nodeProperties(properties);

  var state = transform.state;
  var _state9 = state;
  var document = _state9.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);

  node = node.merge(properties);
  document = document.updateDescendant(node);
  document = document.normalize();
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'set-node',
    path: path,
    properties: properties
  });

  return transform;
}

/**
 * Split a node by `key` at `offset`.
 *
 * @param {Transform} transform
 * @param {String} key
 * @param {Number} offset
 * @return {Transform}
 */

function splitNodeByKey(transform, key, offset) {
  var state = transform.state;
  var _state10 = state;
  var document = _state10.document;

  var node = document.assertDescendant(key);
  var path = document.getPath(node);
  var parent = document.getParent(node);
  var isParent = document == parent;
  var index = parent.nodes.indexOf(node);

  var child = node;
  var one = void 0;
  var two = void 0;

  if (node.kind != 'text') {
    child = node.getTextAtOffset(offset);
  }

  while (child && child != parent) {
    if (child.kind == 'text') {
      var i = node.kind == 'text' ? offset : offset - node.getOffset(child);
      var _child = child;
      var characters = _child.characters;

      var oneChars = characters.take(i);
      var twoChars = characters.skip(i);
      one = child.merge({ characters: oneChars });
      two = child.merge({ characters: twoChars, key: (0, _uid2.default)() });
    } else {
      var _child2 = child;
      var nodes = _child2.nodes;

      var oneNodes = nodes.takeUntil(function (n) {
        return n.key == one.key;
      }).push(one);
      var twoNodes = nodes.skipUntil(function (n) {
        return n.key == one.key;
      }).rest().unshift(two);
      one = child.merge({ nodes: oneNodes });
      two = child.merge({ nodes: twoNodes, key: (0, _uid2.default)() });
    }

    child = document.getParent(child);
  }

  parent = parent.removeNode(index);
  parent = parent.insertNode(index, two);
  parent = parent.insertNode(index, one);
  document = isParent ? parent : document.updateDescendant(parent);
  state = state.merge({ document: document });

  transform.state = state;
  transform.operations.push({
    type: 'split-node',
    offset: offset,
    path: path
  });

  return transform;
}