'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _block = require('../models/block');

var _block2 = _interopRequireDefault(_block);

var _document = require('../models/document');

var _document2 = _interopRequireDefault(_document);

var _inline = require('../models/inline');

var _inline2 = _interopRequireDefault(_inline);

var _data = require('../models/data');

var _data2 = _interopRequireDefault(_data);

var _mark = require('../models/mark');

var _mark2 = _interopRequireDefault(_mark);

var _text = require('../models/text');

var _text2 = _interopRequireDefault(_text);

var _typeOf = require('type-of');

var _typeOf2 = _interopRequireDefault(_typeOf);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Normalize a block argument `value`.
 *
 * @param {Block || String || Object} value
 * @return {Block}
 */

function block(value) {
  if (value instanceof _block2.default) return value;
  var type = (0, _typeOf2.default)(value);

  switch (type) {
    case 'string':
    case 'object':
      {
        return _block2.default.create(nodeProperties(value));
      }
    default:
      {
        throw new Error('Invalid `block` argument! It must be a block, an object, or a string. You passed: "' + value + '".');
      }
  }
}

/**
 * Normalize an inline argument `value`.
 *
 * @param {Inline || String || Object} value
 * @return {Inline}
 */

function inline(value) {
  if (value instanceof _inline2.default) return value;
  var type = (0, _typeOf2.default)(value);

  switch (type) {
    case 'string':
    case 'object':
      {
        return _inline2.default.create(nodeProperties(value));
      }
    default:
      {
        throw new Error('Invalid `inline` argument! It must be an inline, an object, or a string. You passed: "' + value + '".');
      }
  }
}

/**
 * Normalize a key argument `value`.
 *
 * @param {String || Node} value
 * @return {String}
 */

function key(value) {
  if (value instanceof _block2.default) return value.key;
  if (value instanceof _document2.default) return value.key;
  if (value instanceof _inline2.default) return value.key;
  if (value instanceof _text2.default) return value.key;

  var type = (0, _typeOf2.default)(value);
  if (type == 'string') return value;

  throw new Error('Invalid `key` argument! It must be either a block, an inline, a text, or a string. You passed: "' + value + '".');
}

/**
 * Normalize a mark argument `value`.
 *
 * @param {Mark || String || Object} mark
 * @return {Mark}
 */

function mark(value) {
  if (value instanceof _mark2.default) return value;
  var type = (0, _typeOf2.default)(value);

  switch (type) {
    case 'string':
    case 'object':
      {
        return _mark2.default.create(markProperties(value));
      }
    default:
      {
        throw new Error('Invalid `mark` argument! It must be a mark, an object, or a string. You passed: "' + value + '".');
      }
  }
}

/**
 * Normalize a mark properties argument `value`.
 *
 * @param {String || Object} value
 * @return {Object}
 */

function markProperties() {
  var value = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  var ret = {};
  var type = (0, _typeOf2.default)(value);

  switch (type) {
    case 'string':
      {
        ret.type = value;
        break;
      }
    case 'object':
      {
        for (var k in value) {
          if (k == 'data') {
            if (value[k] != null) ret[k] = _data2.default.create(value[k]);
          } else {
            ret[k] = value[k];
          }
        }
        break;
      }
    default:
      {
        throw new Error('Invalid mark `properties` argument! It must be an object or a string. You passed: "' + value + '".');
      }
  }

  return ret;
}

/**
 * Normalize a node properties argument `value`.
 *
 * @param {String || Object} value
 * @return {Object}
 */

function nodeProperties() {
  var value = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  var ret = {};
  var type = (0, _typeOf2.default)(value);

  switch (type) {
    case 'string':
      {
        ret.type = value;
        break;
      }
    case 'object':
      {
        if (value.isVoid != null) ret.isVoid = !!value.isVoid;
        for (var k in value) {
          if (k == 'data') {
            if (value[k] != null) ret[k] = _data2.default.create(value[k]);
          } else {
            ret[k] = value[k];
          }
        }
        break;
      }
    default:
      {
        throw new Error('Invalid node `properties` argument! It must be an object or a string. You passed: "' + value + '".');
      }
  }

  return ret;
}

/**
 * Export.
 *
 * @type {Object}
 */

exports.default = {
  block: block,
  inline: inline,
  key: key,
  mark: mark,
  markProperties: markProperties,
  nodeProperties: nodeProperties
};